// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

library dart.pkg.isolate.loadbalancer_test;

import "package:isolate/loadbalancer.dart";
import "package:isolate/isolaterunner.dart";
import "package:unittest/unittest.dart";
import "dart:async";

main() {
  setGlobal(-1);
  group("4-isolate-lb", () {
    LoadBalancer lb;
    setUp(() {
      // Create a load-balancer with four isolates-runners.
      return Future.wait(new Iterable.generate(4, (_) => IsolateRunner.spawn()))
                   .then((isolates) {
                      return Future.wait([isolates[0].run(setGlobal, 0),
                                          isolates[1].run(setGlobal, 1),
                                          isolates[2].run(setGlobal, 2),
                                          isolates[3].run(setGlobal, 3)])
                                   .then((_) => isolates);
                    })
                    .then((isolates) {
                      lb = new LoadBalancer(isolates);
                    });

    });

    tearDown(() {
      lb.close();
    });

    test("run once", () {
      return lb.run(getGlobal, null).then((v) {
        expect(v, inInclusiveRange(0, 3));
      });
    });

    test("run multiple - 2", () {
      return Future.wait(lb.runMultiple(2, getGlobal, null)).then((vs) {
        var set = vs.toSet();
        expect(set, hasLength(2));
        set.forEach((v) {
          expect(v, inInclusiveRange(0, 3));
        });
      });
    });

    test("run multiple - 4", () {
      return Future.wait(lb.runMultiple(4, getGlobal, null)).then((vs) {
        expect(vs, unorderedMatches([0, 1, 2, 3]));
      });
    });

    test("run multiple - too many", () {
      expect(() => lb.runMultiple(5, getGlobal, null), throws);
    });

    test("load", () {
      return Future.wait([
        lb.run(getGlobal, null, load: 300),
        lb.run(getGlobal, null, load: 305),
        lb.run(getGlobal, null, load: 310),
        lb.run(getGlobal, null, load: 100),
        lb.run(getGlobal, null, load: 100),
        lb.run(getGlobal, null, load: 105),
        lb.run(getGlobal, null, load: 100),
      ]).then((results) {
        int id1 = results[0];
        int id2 = results[1];
        int id3 = results[2];
        int id4 = results[3];
        expect([id1, id2, id3, id4], unorderedMatches([0, 1, 2, 3]));
        expect(results[4], equals(id4));
        expect(results[5], equals(id4));
        expect(results[6], equals(id1));
      });
    });

    test("timeout", () {
      return lb.run(delay, MS * 500, timeout: MS * 100, onTimeout: () => 42)
               .then((v) { expect(v, 42); });
    });

    test("timeout-throws", () {
      return lb.run(delay, MS * 500, timeout: MS * 100)
               .then((v) { fail("unreachable"); },
                     onError: (e) {
                       expect(e is TimeoutException, isTrue);
                     });
    });

    test("multiple-timeout", () {
      return Future.wait(lb.runMultiple(3, delay, MS * 500,
                                        timeout: MS * 100, onTimeout: () => 42))
               .then((vs) {
                 expect(vs, [42, 42, 42]);
               });
    });

  });

}

const MS = const Duration(milliseconds: 1);

var _global;
void setGlobal(v) { _global = v; }
getGlobal(_) => _global;

Future delay(Duration time) => new Future.delayed(time, () => null);
